import torch
import torch.nn as nn
from EIVArchitectures.Layers import EIVInput, EIVDropout, EIVVariationalDropout
from EIVGeneral.repetition import repeat_tensors, reshape_to_chunks

class FNNEIV(nn.Module):
    """
    A fully connected net with Error-in-Variables input and Bernoulli dropout
    layers. 
    :param p: dropout rate, defaults to 0.5
    :param init_std_y: Initial estimated standard deviation for y.
    :param precision_prior_zeta: precision of the prior for zeta.
    Defaults to 0.0 (=improper prior)
    :param deming: Will be used as a coupling factor between std_y and std_x
    (the Deming regression), that is std_x = deming * std_y. 
    :param h: A list specifying the number of neurons in each layer.
    **Note**: 
    To change the deming factor afterwards, use the method `change_deming`
    """
    def __init__(self, p = 0.5, init_std_y=1.0, precision_prior_zeta=0.0, 
            deming=1.0, h=[1, 50,100,50, 1]):
        super().__init__()
        # part before Bernoulli dropout
        self.init_std_y = init_std_y
        InverseSoftplus = lambda sigma: torch.log(torch.exp(sigma) - 1 )
        self.std_y_par = nn.Parameter(
                InverseSoftplus(torch.tensor([init_std_y])))
        self._repetition = 1
        self.main = nn.Sequential(
                EIVInput(precision_prior_zeta=precision_prior_zeta, 
                    external_std_x=self.get_std_x),
                nn.Linear(h[0], h[1]),
                nn.LeakyReLU(1e-2),
                EIVDropout(p=p, repetition_map=self._repetition_map),
                nn.Linear(h[1],h[2]),
                nn.LeakyReLU(1e-2),
                EIVDropout(p=p, repetition_map=self._repetition_map),
                nn.Linear(h[2],h[3]),
                nn.LeakyReLU(1e-2),
                EIVDropout(p=p, repetition_map=self._repetition_map),
                nn.Linear(h[3],h[4]))
        self.p = p
        self._deming = deming
        # needed for switch_noise_off()
        self._stored_deming = deming
        self.noise_is_on = True


    def change_deming(self, deming):
        """
        Update deming factor to `deming`
        :param deming: A positive float
        """
        print('Updating deming from %.3f to %.3f' % (self._deming, deming))
        self._deming = deming
        self._stored_deming = deming


    def noise_off(self):
        self._deming = 0.0
        self.noise_is_on = False

    def noise_on(self):
        self._deming = self._stored_deming
        self.noise_is_on = True

    def sigma(self, y):
        scalar_sigma = self.get_std_y()
        return scalar_sigma.repeat(y.shape)

    def get_std_x(self):
        return self._deming * self.get_std_y()

    def get_std_y(self):
        return nn.Softplus()(self.std_y_par)

    def _repetition_map(self):
        return self._repetition

    def forward(self, x, repetition=1):
        self._repetition = repetition
        mu = self.main(x)
        sigma = self.sigma(mu)
        self._repetition = 1
        return mu, sigma

    def regularizer(self, x, lamb):
        """
        Regularization for EIV net: prior KL term, 
        from "Bernoulli Dropout" by Gal et al., plus the regularizer term 
        from the EIV model (which is constant if precision_prior_zeta is 0).
        :param x: A torch.tensor, the input
        :param lamb: float, prefactor for regularization
        """
        regularization = 0
        p = torch.tensor(self.p)
        last_Dropout_layer = None
        for i, layer in enumerate(self.main):
            if type(layer) is not EIVDropout and last_Dropout_layer != i-1:
                for par in layer.parameters():
                    regularization += lamb*(par**2).sum().view((-1,))
            elif type(layer) is EIVDropout:
                next_layer = self.main[i+1]
                assert type(next_layer) is nn.Linear
                regularization += lamb*(next_layer.bias**2).sum().view((-1,))
                regularization += lamb/(1-p) \
                        * (next_layer.weight**2).sum().view((1,))
                last_Dropout_layer = i
            else:
                pass
        # entropy actually not needed here, added for completeness
        entropy = -1 * (p * torch.log(p) + (1-p) * torch.log(1-p))
        regularization += entropy
        if self._deming > 0:
            # add EIV regularization term 
            # (constant if precision_prior_zeta is 0)
            regularization += self.main[0].neg_x_evidence(x)
        return regularization 


    def predict(self, x, number_of_draws=100, remove_graph=True
            , take_average_of_prediction=True):
        """
        Average over `number_of_draws` forward passes. If 
        `take_average_of_prediction` is False, the averaging is skipped and
        all forward passes are returned.
        **Note**: This method does neither touch the input noise nor Dropout.
        The corresponding setting is left to the user!
        :param x: A torch.tensor, the input
        :param number_of_draws: Number of draws to obtain from x
        :param remove_graph: If True (default) the output will 
        be detached to save memory
        :param take_average_of_prediction: If False, no averaging will be 
        applied to the prediction and the second dimension of the first output
        will count the number_of_draws.
        """
        x, = repeat_tensors(x, number_of_draws=number_of_draws)
        pred, sigma = self.forward(x)
        if remove_graph:
            pred, sigma = pred.detach(), sigma.detach()
        pred, sigma = reshape_to_chunks(pred, sigma, 
                number_of_draws=number_of_draws)
        # reduce along the draws (actually redundant for sigma)
        if take_average_of_prediction:
            pred, sigma = torch.mean(pred, dim=1), torch.mean(sigma, dim=1)
        else: 
            sigma = torch.mean(sigma, dim=1)
        return pred, sigma

class FNNBer(nn.Module):
    """
    A fully connected net Bernoulli dropout layers.
    :param p: dropout rate, defaults to 0.5
    :param init_std_y: Initial standard deviation for input y. 
    :param h: A list specifying the number of neurons in each layer.
    """
    def __init__(self, p=0.5, init_std_y=1.0, h=[1, 50,100,50, 1]):
        super().__init__()
        # part before Bernoulli dropout
        self.init_std_y = init_std_y
        InverseSoftplus = lambda sigma: torch.log(torch.exp(sigma) - 1 )
        self.std_y_par = nn.Parameter(
                InverseSoftplus(torch.tensor([init_std_y])))
        self.main = nn.Sequential(
                nn.Linear(h[0], h[1]),
                nn.LeakyReLU(1e-2),
                nn.Dropout(p=p),
                nn.Linear(h[1],h[2]),
                nn.LeakyReLU(1e-2),
                nn.Dropout(p=p),
                nn.Linear(h[2],h[3]),
                nn.LeakyReLU(1e-2),
                nn.Dropout(p=p),
                nn.Linear(h[3],h[4]))
        self.p = p

    def sigma(self, y):
        scalar_sigma = self.get_std_y()
        return scalar_sigma.repeat(y.shape)

    def get_std_y(self):
        return nn.Softplus()(self.std_y_par)

    def forward(self, x):
        mu = self.main(x)
        sigma = self.sigma(mu)
        return mu, sigma

    def regularizer(self, x, lamb):
        """
        Regularization (prior KL term), from "Bernoulli Dropout" by Gal et al.
        :param x: A torch.tensor, the input
        :param lamb: float, prefactor for regularization
        """
        regularization = 0
        p = torch.tensor(self.p)
        last_Dropout_layer = None
        for i, layer in enumerate(self.main):
            if type(layer) is not nn.Dropout and last_Dropout_layer != i-1:
                for par in layer.parameters():
                    regularization += lamb*(par**2).sum().view((-1,))
            elif type(layer) is nn.Dropout:
                next_layer = self.main[i+1]
                assert type(next_layer) is nn.Linear
                regularization += lamb*(next_layer.bias**2).sum().view((-1,))
                regularization += lamb/(1-p) \
                        * (next_layer.weight**2).sum().view((1,))
                last_Dropout_layer = i
            else:
                pass
        # entropy actually not needed here, added for completeness
        entropy = -1 * (p * torch.log(p) + (1-p) * torch.log(1-p))
        regularization += entropy
        return regularization 


    def predict(self, x, number_of_draws=100, remove_graph=True,
            take_average_of_prediction=True):
        """
        Average over `number_of_draws` forward passes. If 
        `take_average_of_prediction` is False, the averaging is skipped and
        all forward passes are returned.
        **Note**: This method does not touch the Dropout.
        The corresponding setting is left to the user! (analogous to
        the corresponding method for FNNEIV)
        :param x: A torch.tensor, the input
        :param number_of_draws: Number of draws to obtain from x
        :param remove_graph: If True (default) the output will 
        be detached to save memory
        :param take_average_of_prediction: If False, no averaging will be 
        applied to the prediction and the second dimension of the first output
        will count the number_of_draws.
        """
        x, = repeat_tensors(x, number_of_draws=number_of_draws)
        pred, sigma = self.forward(x)
        if remove_graph:
            pred, sigma = pred.detach(), sigma.detach()
        pred, sigma = reshape_to_chunks(pred, sigma, 
                number_of_draws=number_of_draws)
        # reduce along the draws (actually redundant for sigma)
        if take_average_of_prediction:
            pred, sigma = torch.mean(pred, dim=1), torch.mean(sigma, dim=1)
        else: 
            sigma = torch.mean(sigma, dim=1)
        return pred, sigma

class FNN_VD_EIV(nn.Module):
    """
    A fully connected net with Error-in-Variables input and variational dropout
    layers. 
    :param initial_alpha: initial value for the alpha of each VD layer
    :param init_std_y: Initial estimated standard deviation for y.
    :param precision_prior_zeta: precision of the prior for zeta.
    Defaults to 0.0 (=improper prior)
    :param deming: Will be used as a coupling factor between std_y and std_x
    (the Deming regression), that is std_x = deming * std_y. 
    :param h: A list specifying the number of neurons in each layer.
    **Note**: 
    To change the deming factor afterwards, use the method `change_deming`
    """
    def __init__(self, initial_alpha = 0.5, init_std_y=1.0, precision_prior_zeta=0.0, 
            deming=1.0, h=[1, 50,100,50, 1]):
        super().__init__()
        # part before Bernoulli dropout
        self.init_std_y = init_std_y
        InverseSoftplus = lambda sigma: torch.log(torch.exp(sigma) - 1 )
        self.std_y_par = nn.Parameter(
                InverseSoftplus(torch.tensor([init_std_y])))
        self._repetition = 1
        self.main = nn.Sequential(
                EIVInput(precision_prior_zeta=precision_prior_zeta, 
                    external_std_x=self.get_std_x),
                nn.Linear(h[0], h[1]),
                nn.LeakyReLU(1e-2),
                EIVVariationalDropout(initial_alpha=initial_alpha, repetition_map=self._repetition_map),
                nn.Linear(h[1],h[2]),
                nn.LeakyReLU(1e-2),
                EIVVariationalDropout(initial_alpha=initial_alpha, repetition_map=self._repetition_map),
                nn.Linear(h[2],h[3]),
                nn.LeakyReLU(1e-2),
                EIVVariationalDropout(initial_alpha=initial_alpha, repetition_map=self._repetition_map),
                nn.Linear(h[3],h[4]))
        self.initial_alpha = initial_alpha
        self._deming = deming
        # needed for switch_noise_off()
        self._stored_deming = deming
        self.noise_is_on = True


    def change_deming(self, deming):
        """
        Update deming factor to `deming`
        :param deming: A positive float
        """
        print('Updating deming from %.3f to %.3f' % (self._deming, deming))
        self._deming = deming
        self._stored_deming = deming


    def noise_off(self):
        self._deming = 0.0
        self.noise_is_on = False

    def noise_on(self):
        self._deming = self._stored_deming
        self.noise_is_on = True

    def sigma(self, y):
        scalar_sigma = self.get_std_y()
        return scalar_sigma.repeat(y.shape)

    def get_std_x(self):
        return self._deming * self.get_std_y()

    def get_std_y(self):
        return nn.Softplus()(self.std_y_par)

    def _repetition_map(self):
        return self._repetition

    def forward(self, x, repetition=1):
        self._repetition = repetition
        mu = self.main(x)
        sigma = self.sigma(mu)
        self._repetition = 1
        return mu, sigma

    def regularizer(self, x, lamb):
        """
        Regularization for EIV net: prior KL term, 
        from "Bernoulli Dropout" by Gal et al., plus the regularizer term 
        from the EIV model (which is constant if precision_prior_zeta is 0).
        :param x: A torch.tensor, the input
        :param lamb: A list of floats, prefactors for bias regularization
        (first term) and VD-KL term (seconnd term)
        """
        regularization = 0
        last_Dropout_layer = None
        assert type(lamb) is list
        lamb_bias, lamb_kl = lamb
        for i, layer in enumerate(self.main):
            if type(layer) is not EIVVariationalDropout and last_Dropout_layer != i-1:
                for par in layer.parameters():
                    regularization += lamb_bias*(par**2).sum().view((-1,))
            elif type(layer) is EIVVariationalDropout:
                next_layer = self.main[i+1]
                assert type(next_layer) is nn.Linear
                regularization += lamb_bias *(next_layer.bias**2).sum().view((-1,))
                regularization += lamb_kl *\
                    layer.variational_dropout_regularizer()
                last_Dropout_layer = i
            else:
                pass
        if self._deming > 0:
            # add EIV regularization term 
            # (constant if precision_prior_zeta is 0)
            regularization += self.main[0].neg_x_evidence(x)
        return regularization 


    def predict(self, x, number_of_draws=100, remove_graph=True
            , take_average_of_prediction=True):
        """
        Average over `number_of_draws` forward passes. If 
        `take_average_of_prediction` is False, the averaging is skipped and
        all forward passes are returned.
        **Note**: This method does neither touch the input noise nor Dropout.
        The corresponding setting is left to the user!
        :param x: A torch.tensor, the input
        :param number_of_draws: Number of draws to obtain from x
        :param remove_graph: If True (default) the output will 
        be detached to save memory
        :param take_average_of_prediction: If False, no averaging will be 
        applied to the prediction and the second dimension of the first output
        will count the number_of_draws.
        """
        x, = repeat_tensors(x, number_of_draws=number_of_draws)
        pred, sigma = self.forward(x)
        if remove_graph:
            pred, sigma = pred.detach(), sigma.detach()
        pred, sigma = reshape_to_chunks(pred, sigma, 
                number_of_draws=number_of_draws)
        # reduce along the draws (actually redundant for sigma)
        if take_average_of_prediction:
            pred, sigma = torch.mean(pred, dim=1), torch.mean(sigma, dim=1)
        else: 
            sigma = torch.mean(sigma, dim=1)
        return pred, sigma

class FNN_VD_Ber(nn.Module):
    """
    A fully connected net with Variational dropout layers.
    :param initial_alpha: initial value for the alpha of each VD layer
    :param init_std_y: Initial standard deviation for input y. 
    :param h: A list specifying the number of neurons in each layer.
    """
    def __init__(self, initial_alpha=0.5, init_std_y=1.0, h=[1, 50,100,50, 1]):
        super().__init__()
        # part before Bernoulli dropout
        self.init_std_y = init_std_y
        InverseSoftplus = lambda sigma: torch.log(torch.exp(sigma) - 1 )
        self.std_y_par = nn.Parameter(
                InverseSoftplus(torch.tensor([init_std_y])))
        self.main = nn.Sequential(
                nn.Linear(h[0], h[1]),
                nn.LeakyReLU(1e-2),
                EIVVariationalDropout(initial_alpha=initial_alpha),
                nn.Linear(h[1],h[2]),
                nn.LeakyReLU(1e-2),
                EIVVariationalDropout(initial_alpha=initial_alpha),
                nn.Linear(h[2],h[3]),
                nn.LeakyReLU(1e-2),
                EIVVariationalDropout(initial_alpha=initial_alpha),
                nn.Linear(h[3],h[4]))
        self.initial_alpha = initial_alpha

    def sigma(self, y):
        scalar_sigma = self.get_std_y()
        return scalar_sigma.repeat(y.shape)

    def get_std_y(self):
        return nn.Softplus()(self.std_y_par)

    def forward(self, x):
        mu = self.main(x)
        sigma = self.sigma(mu)
        return mu, sigma

    def regularizer(self, x, lamb):
        """
        Regularization (prior KL term), from "Bernoulli Dropout" by Gal et al.
        :param x: A torch.tensor, the input
        :param lamb: float, prefactor for regularization
        """
        regularization = 0
        last_Dropout_layer = None
        assert type(lamb) is list
        lamb_bias, lamb_kl = lamb
        for i, layer in enumerate(self.main):
            if type(layer) is not EIVVariationalDropout and last_Dropout_layer != i-1:
                for par in layer.parameters():
                    regularization += lamb_bias * (par**2).sum().view((-1,))
            elif type(layer) is EIVVariationalDropout:
                next_layer = self.main[i+1]
                assert type(next_layer) is nn.Linear
                regularization += lamb_bias *\
                    (next_layer.bias**2).sum().view((-1,))
                regularization += lamb_kl *\
                        layer.variational_dropout_regularizer()
                last_Dropout_layer = i
            else:
                pass
        return regularization 

    def predict(self, x, number_of_draws=100, remove_graph=True,
            take_average_of_prediction=True):
        """
        Average over `number_of_draws` forward passes. If 
        `take_average_of_prediction` is False, the averaging is skipped and
        all forward passes are returned.
        **Note**: This method does not touch the Dropout.
        The corresponding setting is left to the user! (analogous to
        the corresponding method for FNNEIV)
        :param x: A torch.tensor, the input
        :param number_of_draws: Number of draws to obtain from x
        :param remove_graph: If True (default) the output will 
        be detached to save memory
        :param take_average_of_prediction: If False, no averaging will be 
        applied to the prediction and the second dimension of the first output
        will count the number_of_draws.
        """
        x, = repeat_tensors(x, number_of_draws=number_of_draws)
        pred, sigma = self.forward(x)
        if remove_graph:
            pred, sigma = pred.detach(), sigma.detach()
        pred, sigma = reshape_to_chunks(pred, sigma, 
                number_of_draws=number_of_draws)
        # reduce along the draws (actually redundant for sigma)
        if take_average_of_prediction:
            pred, sigma = torch.mean(pred, dim=1), torch.mean(sigma, dim=1)
        else: 
            sigma = torch.mean(sigma, dim=1)
        return pred, sigma

